import sys
from asciimatics.effects import Background, Cycle, Stars, Print, Sprite
from asciimatics.renderers import FigletText, Box, StaticRenderer, SpeechBubble
from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.exceptions import ResizeScreenError
from asciimatics.paths import DynamicPath
from asciimatics.event import KeyboardEvent, MouseEvent
from .game import *
from .ai_player import *

symbol_placer = None
current_turn = 0
bigBoxLength = 0
pvp = False

game = None

class MouseController(DynamicPath):
    def __init__(self, sprite, scene, x, y):
        super(MouseController, self).__init__(scene, x, y)
        self._sprite = sprite

    def process_event(self, event):
        if isinstance(event, MouseEvent):
            self._x = event.x
            self._y = event.y
            if event.buttons & MouseEvent.LEFT_CLICK != 0:
                # Try to whack the other sprites when mouse is clicked
                self._sprite.place_symbol()
        else:
            return event

class SymbolPlacer(Sprite):
    def __init__(self, screen):
        """
        See :py:obj:`.Sprite` for details.
        """
        super(SymbolPlacer, self).__init__(
            screen,
            renderer_dict={
                "default": StaticRenderer(images=["X"])
            },
            path=MouseController(
                self, screen, screen.width // 2, screen.height // 2),
            colour=Screen.COLOUR_RED)

    def place_symbol(self):
        global pvp

        if game.get_game_over():
            return

        x, y = self._path.next_pos()
        symbolX, symbolY = 0, 0
        boxOriginX = int(self._screen.width / 2 - bigBoxLength / 2 * 2)
        boxOriginY = int(self._screen.height / 2 - bigBoxLength / 2)
        validPos = True

        this_turn_board_row = 0
        this_turn_board_col = 0
        if x > boxOriginX and x < boxOriginX + bigBoxLength * 2:
            if x < int(boxOriginX + bigBoxLength / 3 * 2):
                this_turn_board_col = 0
            elif x > int(boxOriginX + bigBoxLength * 2 / 3 * 2):
                this_turn_board_col = 2
            else:
                this_turn_board_col = 1
        else:
            validPos = False
        
        if y > boxOriginY and y < boxOriginY + bigBoxLength:
            if y < int(boxOriginY + bigBoxLength / 3):
                this_turn_board_row = 0
            elif y > int(boxOriginY + bigBoxLength * 2 / 3):
                this_turn_board_row = 2
            else:
                this_turn_board_row = 1
        else:
            validPos = False

        selected_square = 3 * (this_turn_board_row) + this_turn_board_col
        if validPos and game.player_turn(selected_square):
            if not game.get_game_over() and not pvp:
                game.ai_player_turn()

            self.print_board()

        if game.check_win() != "":
            self._scene.add_effect(Print(self._screen,
            FigletText("Winner is: " + str(game.check_win())),
            int(bigBoxLength * 9 / 8), transparent=False))
        elif game.get_game_over():
            self._scene.add_effect(Print(self._screen,
            FigletText("Nobody wins"),
            int(bigBoxLength * 9 / 8), transparent=False))

    def print_board(self):
        boxOriginX = int(self._screen.width / 2 - bigBoxLength / 2 * 2)
        boxOriginY = int(self._screen.height / 2 - bigBoxLength / 2)
        symbolX, symbolY = 0, 0

        for i in range(9):
            # get screen pos of col
            if i % 3 == 0:
                symbolX = int(boxOriginX + bigBoxLength / 6 * 2)
            elif i % 3 == 1:
                symbolX = int(boxOriginX + bigBoxLength / 2 * 2)
            elif i % 3 == 2:
                symbolX = int(boxOriginX + bigBoxLength * 5 / 6 * 2)

            # get screen pos of row
            if i / 3 == 0:
                symbolY = int(boxOriginY + bigBoxLength / 6)
            elif i / 3 == 1:
                symbolY = int(boxOriginY + bigBoxLength / 2)
            elif i / 3 == 2:
                symbolY = int(boxOriginY + bigBoxLength * 5 / 6)

            self._scene.add_effect(Print(self._screen,
                FigletText(game.board.view_board()[i]),
                symbolY-4, symbolX-2))
        
        

def demo(screen):
    global symbol_placer, bigBoxLength
    symbol_placer = SymbolPlacer(screen)
    bigBoxLength = int(screen.width / 4 * 3) if screen.width < screen.height else int(screen.height / 4 * 3)
    effects = [
        Print(screen,
            Box(bigBoxLength * 2, bigBoxLength),
            int(bigBoxLength / 8)),
        Print(screen,
            FigletText("tictactAI"),
            int(bigBoxLength * 9 / 8)),
        symbol_placer
    ]
    screen.play([Scene(effects, 500)])

if __name__ == "__main__":
    while True:
        try:
            Screen.wrapper(demo)
            sys.exit(0)
        except ResizeScreenError:
            pass

def start(isPvp):
    global pvp, game 
    pvp = isPvp
    game = Game(playerO = AI_Player()) if not pvp else Game()
    while True:
        try:
            Screen.wrapper(demo)
            sys.exit(0)
        except ResizeScreenError:
            pass


from .player import *
from .evaluator import *

class AI_Player(Player):
    def __init__(self, symbol="O"):
        super().__init__(self)
        self.__symbol = symbol

    def place_symbol_on_board(self, board) -> int:
        bestMove = get_best_move(board, False)
        board.place_symbol(bestMove, self.__symbol)
        return bestMove

    def get_symbol(self) -> str:
        return self.__symbol



import click
from .output import *

@click.command()
@click.option('--pvp/--ai', default=False, help='Play vs a player or AI')
def cli(pvp):
    """Play a game of tic tac toe against an AI"""
    start(pvp)


from .player import *
from .ai_player import *
from .evaluator import *

class Game():
    def __init__(self, playerX=Player("X"), playerO=Player("O")):
        self.winner = None
        self.playerX = playerX
        self.playerO = playerO
        self.board = Board()
        self.__turn_counter = 0
        self.__current_player = playerX
        self.__game_over = False

    def get_turn(self) -> int:
        return self.__turn_counter

    def get_current_player(self) -> Player:
        return self.__current_player

    def player_turn(self, position_index) -> bool:
        if not self.__game_over:
            self.__current_player = self.playerX if self.__turn_counter % 2 == 0 else self.playerO
            successfully_place_symbol = self.__current_player.place_symbol_on_board(position_index, self.board)
            self.__turn_counter += 1 if successfully_place_symbol else 0
            self.__game_over = has_winner(self.board.view_board()) or has_draw(self.board.view_board())
            return successfully_place_symbol
        else: return False
        
    def ai_player_turn(self) -> int:
        if not self.__game_over:
            self.__current_player = self.playerX if self.__turn_counter % 2 == 0 else self.playerO
            bestMove = self.playerO.place_symbol_on_board(self.board)
            self.__turn_counter += 1
            self.__game_over = has_winner(self.board.view_board()) or has_draw(self.board.view_board())
            return bestMove

    def get_game_over(self) -> bool:
        return self.__game_over

    def check_win(self) -> str:
        winner = get_winner(self.board.view_board())
        self.__game_over = has_winner(self.board.view_board()) or has_draw(self.board.view_board())
        return winner

    
    



import sys
sys.path.append('../src')
from .board import *

def get_winner(board) -> str:
    row_winner = check_row_win(board)
    col_winner = check_col_win(board)
    diagonal_winner = check_diagonal_win(board)
    if not row_winner == "":
        return row_winner
    elif not col_winner == "":
        return col_winner
    elif not diagonal_winner == "":
        return diagonal_winner
    else:
        return ""

def has_winner(board) -> bool:
    return get_winner(board) != ""

def has_draw(board) -> bool:
    if not has_winner(board) and "" not in board:
        return True
    else: 
        return False

def check_row_win(board) -> str:
    x_count = 0
    o_count = 0
    winner = ""

    for row in range(3):
        x_count = 0
        o_count = 0

        for col in range(3):
            if board[row * 3 + col] == "X":
                x_count +=  1
            elif board[row * 3 + col] == "O":
                o_count += 1
            if x_count == 3: 
                winner = "X"
            elif o_count == 3: 
                winner = "O"
    
    return winner

def check_col_win(board) -> str:
    x_count = 0
    o_count = 0
    winner = ""

    for col in range(3):
        x_count = 0
        o_count = 0

        for row in range(3):
            if board[row * 3 + col] == "X":
                x_count += 1
            elif board[row * 3 + col] == "O":
                o_count += 1
            if x_count == 3:
                winner = "X"
            elif o_count == 3:
                winner = "O"

    return winner
        
def check_diagonal_win(board) -> str:
    if board[0] == "X" and board[4] == "X" and board[8] == "X":
        return "X"
    elif board[0] == "O" and board[4] == "O" and board[8] == "O":
        return "O"
    elif board[2] == "X" and board[4] == "X" and board[6] == "X":
        return "X"
    elif board[2] == "O" and board[4] == "O" and board[6] == "O":
        return "O"

    return ""
    
# X is maximizer, O is minimizer
def evaluate_board(board) -> int:
    if get_winner(board) == "X": return 10
    elif get_winner(board) == "O": return -10
    elif has_draw(board): return 0

# considers all possible game outcomes and returns the value of the board
def minimax(boardObj, isMaximizingPlayer, depth=0, alpha=-1000, beta=1000):
    score = evaluate_board(boardObj.view_board())

    # game ended with winner
    if score == 10:
        return score - depth

    if score == -10:
        return score + depth

    if has_draw(boardObj.view_board()):
        return 0

    # player X
    if isMaximizingPlayer:
        bestVal = -1000

        #check each move
        for i in range(9):
            if boardObj.view_board()[i] == "":
                boardObj.place_symbol(i, "X")

                #replace best value by calling minimax recursively on subsequent moves after this one
                bestVal = max(bestVal, minimax(boardObj, not isMaximizingPlayer, depth+1, alpha, beta))
                alpha = max(alpha, bestVal)
                
                #reverse move
                boardObj.remove_symbol(i)

                #alpha-beta pruning
                if beta <= alpha:
                    break

        return bestVal

    # player O, minimizing player
    else:
        bestVal = 1000

        #check each move
        for i in range(9):
            if boardObj.view_board()[i] == "":
                boardObj.place_symbol(i, "O")

                #replace best value by calling minimax recursively on subsequent moves after this one
                bestVal = min(bestVal, minimax(boardObj, not isMaximizingPlayer, depth+1, alpha, beta))
                beta = min(beta, bestVal)
                
                #reverse move
                boardObj.remove_symbol(i)

                #alpha-beta pruning
                if beta <= alpha:
                    break

        return bestVal

# returns best position to place symbol
def get_best_move(boardObj, isMaximizingPlayer) -> int:
    # player X
    if isMaximizingPlayer:
        bestVal = -1000
        bestMove = -1

        #check each move
        for i in range(9):
            if boardObj.view_board()[i] == "":
                boardObj.place_symbol(i, "X")

                #replace best value by calling minimax recursively on subsequent moves after this one
                moveVal = minimax(boardObj, not isMaximizingPlayer)
                
                #reverse move
                boardObj.remove_symbol(i)

                if moveVal > bestVal: 
                    bestVal = moveVal
                    bestMove = i

        return bestMove

    # player O, minimizing player
    else:
        bestVal = 1000
        bestMove = -1

        #check each move
        for i in range(9):
            if boardObj.view_board()[i] == "":
                boardObj.place_symbol(i, "O")

                #replace best value by calling minimax recursively on subsequent moves after this one
                moveVal = minimax(boardObj, not isMaximizingPlayer)
                
                #reverse move
                boardObj.remove_symbol(i)

                if moveVal < bestVal: 
                    bestVal = moveVal
                    bestMove = i

        return bestMove



from .board import *

class Player():
    def __init__(self, symbol="X"):
        self.__symbol = symbol

    def place_symbol_on_board(self, index, board) -> bool:
        return board.place_symbol(index, self.__symbol)

    def get_symbol(self) -> str:
        return self.__symbol


from typing import List

class Board():
    def __init__(self):
        self.__board = ["", "", "",
                        "", "", "",
                        "", "", ""]
                        
    def view_board(self) -> List[str]:
        return self.__board.copy()
    
    def place_symbol(self, index, symbol) -> bool:
        if self.__board[index] == "":
            self.__board[index] = symbol
            return True
        else:
            return False

    def set_board(self, boardList):
        self.__board = boardList

    def remove_symbol(self, index) -> bool:
        if not self.__board[index] == "":
            self.__board[index] = ""
            return True
        else:
            return False

    def __str__(self):
        return self.__board.__str__()



from setuptools import setup, find_packages

from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

readme = open('README.md', 'r')
README_TEXT = readme.read()
readme.close()

setup(
    name='tictactai',
    version='1.0.8',
    license='MIT',
    long_description=README_TEXT,
    long_description_content_type='text/markdown',
    description='Play tic tac toe with an AI',
    author='Daniel Ng',
    author_email='dndanielng@gmail.com',
    url='https://github.com/donutboy929/tictactAI',
    download_url='https://github.com/donutboy929/tictactAI/archive/refs/tags/v1.0.tar.gz',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('*.py')],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    keywords=['tui', 'tictactoe', 'minimax'],
    python_requires='>=3.7',
    classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Topic :: Games/Entertainment :: Board Games',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.9',
    ],
    install_requires=[
        'Click',
        'asciimatics==1.13.0',
        'attrs==21.2.0',
        'future==0.18.2',
        'iniconfig==1.1.1',
        'packaging==21.0',
        'Pillow==8.4.0',
        'pluggy==1.0.0',
        'py==1.10.0',
        'pyfiglet==0.8.post1',
        'pyparsing==3.0.3',
        'pytest==6.2.5',
        'toml==0.10.2',
        'wcwidth==0.2.5',
        'importlib-metadata',
        'typing-extensions',
        'zipp',
    ],entry_points='''
        [console_scripts]
        tictactai=tictactai.tictactai:cli
    '''
)